import random
import warnings
import matplotlib.pyplot as plt
from Finch.exceptions.environment_exceptions import NoIndividualsAtEndOfRun
import time
from tabulate import tabulate
from Finch.layers import layer

from Finch.genetics import Individual


class Environment(layer.Layer):
    """
        Represents an evolutionary environment that manages the evolution of a population of individuals over generations.

        Args:
        - layers (list[layer.Layer]): List of layers in the environment.
        - name (str): Name of the environment.
        - verbose_every (int): Frequency of verbose output during evolution (aka print every n times)
        - device (str): Device to run the environment on. 'gpu' or 'cpu'

        Attributes:
        - fitness_function: The fitness function used for evaluating individuals.
        - generations: Number of generations to evolve for.
        - callback: Callback function to be executed after each generation. Set this in .compile()
        - layers: List of layers in the environment.
        - individuals: List of individuals in the population.
        - iteration: Current iteration during evolution.
        - history: List to store best fitness values over generations.
        - compiled: Indicates whether the environment has been compiled.
        - deactivated: Indicates whether the environment is deactivated.
        - best_ever: The best individual observed during evolution.
        """

    def __init__(self, layers: list[layer.Layer] = None, name="Environment", verbose_every=1, device='cpu'):
        super().__init__(device=device, refit=False, individual_selection=None)
        self.fitness_function = None

        self.generations = 0
        self.verbose_every = verbose_every
        self.callback = None
        self.name = name
        self.layers = layers
        for lay in layers:
            lay.set_environment(self)
        self.individuals = []
        self.iteration = 0
        self.history = []
        self.compiled = False
        self.deactivated = False
        self.best_ever = None
        self.dead_individuals = []

    def deactivate(self):
        """
        Deactivates the environment, preventing further evolution.
        """
        self.deactivated = True

    def get_fitness_metric(self):
        """
        Returns the fitness metric of the best individual observed so far.
        """

        if self.best_ever:
            return self.best_ever.fitness
        else:
            return 0

    def batch(self, individual):
        individual.check_fitness = True
        return individual.fitness

    def compile(self, fitness_function, individuals: list[Individual] = None, callback: callable = None,
                verbose_every: int = 1):
        """
        Compiles the environment with necessary parameters for evolution.

        Args:
        - fitness_function: The fitness function for evaluating individuals.
        - individuals (list[Individual]): List of individuals to start evolution with.
        - callback (callable): Callback function to be executed after each generation.
        - verbose_every (int): Frequency of verbose output during evolution.
        """
        self.fitness_function = fitness_function

        if self.fitness_function == 'batch':
            self.fitness_function = self.batch

        self.individuals = individuals
        self.compiled = True
        if self.individuals is None:
            self.individuals = []
        self.callback = callback
        self.verbose_every = verbose_every

    def evolve(self, generations):
        """
        Evolves the population over a specified number of generations.

        Args:
        - generations (int): Number of generations for evolution.

        Returns:
        Tuple containing the final population and the fitness history.
        """

        if self.deactivated:
            return self.individuals
        self.generations = generations
        for i in range(self.generations):
            self.iteration = i
            self.execute(self.individuals)
        return self.individuals, self.history

    def execute(self, individuals: list[Individual]):
        """
        Executes one generation of evolution.

        Args:
        - individuals (list[Individual]): List of individuals in the population.

        Raises:
        - NoIndividualsAtEndOfRun: If the environment has a population of 0 after running.
        """
        for a_layer in self.layers:
            a_layer.run(self.individuals, self)
        if self.callback:
            self.callback(self.individuals, self)
        if len(self.individuals) == 0:
            raise NoIndividualsAtEndOfRun("Your environment has a population of 0 after running.")
        fitness = self.individuals[0].fitness
        if self.best_ever:
            if fitness > self.best_ever.fitness:
                self.best_ever = self.individuals[0].copy()
        else:
            self.best_ever = self.individuals[0].copy()
        if self.verbose_every and self.iteration % self.verbose_every == 0 and self.iteration > 1:            print(
            f"{self.name}: generation {self.iteration + 1}/{self.generations}. Max fitness: {fitness}. Population: "
            f"{len(self.individuals)}")
        self.history.append(fitness)

    def plot(self):
        plt.title("Fitness graph")
        plt.plot(self.history)
        plt.show()


class Sequential(Environment):
    """
        Represents a sequential evolutionary environment.

        Args:
        - layers: List of layers in the environment.
        - name: Name of the environment.
        """

    def __init__(self, layers, name="default"):
        super().__init__(layers, name)
        self.layers = layers

    def compile(self, fitness_function, individuals: list[Individual] = None, callback: callable = None,
                verbose_every: int = 1):
        super().compile(individuals=individuals, callback=callback, verbose_every=verbose_every,
                        fitness_function=fitness_function)

    def evolve(self, generations: int):
        return super().evolve(generations)

    def reset(self):
        self.individuals = []
        self.history = []
        self.iteration = 0


class Adversarial(Environment):
    """
    Represents an adversarial evolutionary environment managing multiple sub-environments.

    Args:
    - environments: List of sub-environments.
    - name: Name of the adversarial environment.
    """

    def __init__(self, environments, name="Adversarial Environment"):
        super().__init__(environments, name)
        self.environments = environments

    def compile(self, fitness_function, individuals: list[Individual] = None, callback: callable = None,
                verbose_every: int = 1):
        """
         Compiles the adversarial environment with necessary parameters for evolution.

         Args:
         - fitness_function: The fitness function for evaluating individuals.
         - individuals (list[Individual]): List of individuals to start evolution with.
         - callback (callable): Callback function to be executed after each generation.
         - verbose_every (int): Frequency of verbose output during evolution.
        """
        super().compile(individuals=individuals, callback=callback, verbose_every=verbose_every,
                        fitness_function=fitness_function)

        for environment in self.environments:
            if not environment.compiled:
                environment.compile(verbose_every=verbose_every)

    def evolve(self, generations: int = 1):
        """
        Evolves the population over a specified number of generations for each sub-environment.

        Args:
        - generations (int): Number of generations for evolution.

        Returns:
        Tuple containing information about the best-performing sub-environment.
        """
        results = []
        results_history = []
        for env in self.environments:
            start_time = time.time()
            individuals, history = env.evolve(generations)
            end_time = time.time()
            evolution_time = end_time - start_time
            results.append((env.name, history[-1], evolution_time))
            results_history.append((env.name, history[-1], evolution_time, history))
            if self.callback:
                self.callback(individuals)
        table_headers = ["Environment", "Max Fitness", "Evolution Time (s)"]

        print(tabulate(results, headers=table_headers, tablefmt="pretty"))
        best_environment = max(results, key=lambda x: x[1])  # Choose the environment with the highest fitness
        print(f"Best environment: {best_environment[0]}, Max fitness: {best_environment[1]}")

        self.plot_fitness_histories(results_history)  # Call the plot function with results

        return best_environment

    def plot_fitness_histories(self, results):
        """
        Plots the fitness histories of multiple sub-environments.

        Args:
        - results: List containing information about each sub-environment's evolution.
        """
        plt.figure(figsize=(10, 6))
        for env_name, max_fitness, _, history in results:
            plt.plot(history, label=env_name)
        plt.xlabel("Generation")
        plt.ylabel("Fitness")
        plt.title(self.name + ", Fitness History Comparison")
        plt.legend()
        plt.grid(True)
        plt.show()


class ChronologicalEnvironment(Environment):
    """
    Represents a chronological evolutionary environment with varying sub-environments.

    Args:
    - environments_and_generations: List of tuples containing sub-environments and their corresponding generations.
    - name: Name of the chronological environment.
    """

    def __init__(self, environments_and_generations, name="mixed_environment"):
        super().__init__(None, name)
        self.environments_and_generations = environments_and_generations
        self.combined_history = []
        self.best_environment = None
        self.individuals = []

    def evolve(self, generations=1):
        """
        Evolves the population over a specified number of generations, using different sub-environments for each period.

        Args:
        - generations (int): Number of generations for evolution.

        Returns:
        Tuple containing the final population and the combined fitness history.
        """

        for env, generations_here in self.environments_and_generations:
            if generations_here is None:
                warnings.warn("Generally avoid using more than 1 generation in a ChronologicalEnvironment, instead "
                              "pass them in the init")
                generations_here = generations
            print(f"Running {env.name} for {generations_here} generations...")
            env.individuals = self.individuals
            individuals, history = env.evolve(generations_here)
            env.history = []
            self.individuals.extend(individuals)
            self.combined_history.extend(history)
            if not self.best_environment or history[-1] > self.best_environment[1]:
                self.best_environment = (env.name, history[-1])
        return self.individuals, self.combined_history

    def compile(self, fitness_function, individuals: list[Individual] = None, callback: callable = None,
                verbose_every: int = 1):
        """
        Compiles the chronological environment with necessary parameters for evolution.

        Args:
        - fitness_function: The fitness function for evaluating individuals.
        - individuals (list[Individual]): List of individuals to start evolution with.
        - callback (callable): Callback function to be executed after each generation.
        - verbose_every (int): Frequency of verbose output during evolution.
        """
        super().compile(individuals=individuals, callback=callback, verbose_every=verbose_every,
                        fitness_function=fitness_function)

        for environment, generations_here in self.environments_and_generations:
            if not environment.compiled:
                environment.compile(verbose_every=verbose_every)

    def plot(self):
        """
        Plots the combined fitness history of sub-environments over time.
        """
        plt.figure(figsize=(10, 6))

        # Define a custom color palette with distinct colors
        custom_colors = ['#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd', '#8c564b']
        start_idx = 0

        for idx, (env, generations) in enumerate(self.environments_and_generations):
            end_idx = start_idx + generations
            history = self.combined_history[start_idx:end_idx]

            # Use modulo to cycle through custom colors
            color_idx = idx % len(custom_colors)

            # Create a line segment for this environment with a changing color
            plt.plot(range(start_idx + 1, end_idx + 1), history, label=f"{env.name}", color=custom_colors[color_idx])

            start_idx = end_idx

        plt.xlabel("Generation")
        plt.ylabel("Fitness")
        plt.title("Combined Fitness History")
        plt.legend()
        plt.grid(True)
        plt.show()


class AdaptiveEnvironment(Environment):
    """
    Represents an adaptive evolutionary environment that switches between different sub-environments.

    Args:
    - environments: List of sub-environments.
    - switch_every: Frequency of environment switching.
    - go_for: Number of generations to run each sub-environment before switching.
    - name: Name of the adaptive environment.
    """

    def __init__(self, environments, switch_every=10, go_for=0, name="AdaptiveEnvironment"):
        super().__init__(name=name)
        self.environments = environments
        self.switch_every = switch_every
        self.current_environment = environments[0]
        self.current_generation = 0
        self.go_for = go_for

    def compile(self, fitness_function, individuals: list[Individual] = None, callback: callable = None,
                verbose_every: int = 1):
        """
        Compiles the adaptive environment with necessary parameters for evolution.

        Args:
        - fitness_function: The fitness function for evaluating individuals.
        - individuals (list[Individual]): List of individuals to start evolution with.
        - callback (callable): Callback function to be executed after each generation.
        - verbose_every (int): Frequency of verbose output during evolution.
        """

        super().compile(individuals=individuals, callback=callback, verbose_every=verbose_every,
                        fitness_function=fitness_function)

        for environment in self.environments:
            if not environment.compiled:
                environment.compile(verbose_every=verbose_every)

    def evolve(self, generations: int):
        """
        Evolves the population over a specified number of generations with adaptive switching between sub-environments.

        Args:
        - generations (int): Number of generations for evolution.

        Returns:
        Tuple containing the final population and the fitness history.
        """

        for i in range(generations):
            self.iteration += 1
            self.current_generation += 1
            new = self.smart_evolve()
            new.individuals = self.current_environment.individuals
            new.history = self.current_environment.history
            self.current_environment = new
            self.current_environment.evolve(1)

        return self.individuals, self.history

    def smart_evolve(self):
        """
        Implements adaptive evolution by intelligently switching between sub-environments.

        Returns:
        The selected sub-environment for the next generation.
        """
        for i in self.environments:
            if i.fitness < 1:
                print("Deactivating ", i.name)
                i.deactivate()
        if self.iteration % self.switch_every == 0:
            weights = [x.fitness for x in self.environments]
            choices = [x for x in self.environments]
            return random.choices(choices, weights, k=1)[0]
        else:
            return self.current_environment

    def switch_environment(self):
        """
        Switches to the sub-environment with the highest fitness increase.
        """

        max_fitness_increase = -float('inf')
        best_environment = None
        first = self.current_environment.individuals[-1].fitness
        for environment in self.environments:
            environment.individuals = self.current_environment.individuals

            environment.evolve(self.go_for)
            new_fitness = environment.individuals[-1].fitness
            fitness_increase = new_fitness - first

            if fitness_increase > max_fitness_increase:
                max_fitness_increase = fitness_increase
                best_environment = environment

        if best_environment:
            best_environment.individuals = self.current_environment.individuals
            self.current_environment = best_environment
            print(f"Switching to environment: {best_environment.name} (Generation {self.current_generation})")

    def plot(self):
        """
        Plots the fitness history of the current sub-environment.
        """
        plt.plot(self.current_environment.history)
        plt.show()



import numpy as np
import Finch.layers.layer as layer
from typing import Union, Callable
from Finch.tools.rates import make_callable

cp = None
try:
    import cupy as cp
except ImportError:
    pass


class Mutate(layer.Mutate):
    def __init__(self, individual_selection: Union[float, int, Callable], gene_selection: Union[float, int, Callable],
                 refit=True):
        super().__init__(individual_selection=individual_selection, gene_selection=gene_selection,
                         refit=refit)

    def mutate_one(self, individual, environment):
        """
        Mutates a single individual
        :param environment: the environment the individuals live in
        :param individual: the individual to mutate
        :return: None
        """
        idx = self.gene_selection(individual)
        individual.genes[idx] = individual.gene_pool.generate_genes(idx.size)


class FloatMutateRange(layer.Mutate):
    """
    A float based layer that mutates genes by changing their values rather than replacing them.
    This can allow for more fine-grained mutations.
    """

    def __init__(self, individual_selection: Union[float, int, Callable], gene_selection: Union[float, int, Callable],
                 max_mutation: Union[float, int] = 1.0, min_mutation: Union[float, int] = -1.0,
                 refit=True, keep_within_genepool_bounds=False):
        """
        :param max_mutation: maximum mutation
        :param min_mutation: minimum mutation
        :param individual_selection: individual selection method
        :param gene_selection: gene selection method
        :param keep_within_genepool_bounds: keeps the values of the floats within the max
        and min set in their gene pools
        """
        super().__init__(individual_selection=individual_selection, gene_selection=gene_selection,
                         refit=refit)
        self.max_mutation = make_callable(max_mutation)
        self.min_mutation = make_callable(min_mutation)
        self.keep_within_genepool_bounds = keep_within_genepool_bounds

    def mutate_one(self, individual, environment):
        """
        Mutates a single individual
        :param environment: the environment the individuals live in
        :param individual: the individual to mutate
        :return: None
        """
        genes_to_change = self.gene_selection(individual)
        if individual.device == 'gpu':
            npcp = cp
        else:
            npcp = np

        individual.genes[genes_to_change] += npcp.random.uniform(self.max_mutation(), self.min_mutation(),
                                                                 size=genes_to_change.size)
        if self.keep_within_genepool_bounds:
            individual.genes[individual.genes > individual.gene_pool.maximum] = individual.gene_pool.maximum
            individual.genes[individual.genes < individual.gene_pool.minimum] = individual.gene_pool.minimum


class IntMutateRange(FloatMutateRange):
    """
    An integer-based layer that mutates genes by changing their values rather than replacing them.
    This can allow for more fine-grained mutations.
    """

    def __init__(self, gene_selection: Union[float, int, Callable], individual_selection: Union[float, int, Callable],
                 max_mutation: Union[int, Callable] = 1, min_mutation: Union[int, Callable] = -1,
                 refit=True, keep_within_genepool_bounds=False):
        """
        :param max_mutation: maximum mutation
        :param min_mutation: minimum mutation
        :param individual_selection: individual selection method
        :param gene_selection: gene selection method
        :param keep_within_genepool_bounds: keep within bounds of possible genes
        """
        super().__init__(max_mutation=max_mutation, min_mutation=min_mutation,
                         individual_selection=individual_selection,
                         gene_selection=gene_selection,
                         refit=refit, keep_within_genepool_bounds=keep_within_genepool_bounds)

    def mutate_one(self, individual, environment):
        """
        Mutates a single individual
        :param environment: the environment the individuals live in
        :param individual: the individual to mutate
        :return: None
        """
        if individual.device == 'gpu':
            npcp = cp
        else:
            npcp = np

        genes_to_change = self.gene_selection(individual)

        individual.genes[genes_to_change] += npcp.random.randint(self.min_mutation(), self.max_mutation() + 1,
                                                           size=genes_to_change.size)

        if self.keep_within_genepool_bounds:
            individual.genes[individual.genes > individual.gene_pool.maximum] = individual.gene_pool.maximum
            individual.genes[individual.genes < individual.gene_pool.minimum] = individual.gene_pool.minimum


class BinaryMutate(layer.Mutate):
    """
    A binary-based layer that mutates genes by changing their values to 0 or 1.
    """

    def __init__(self, individual_selection: Union[float, int, Callable], gene_selection: Union[float, int, Callable],
                 refit=True):
        super().__init__(individual_selection=individual_selection, gene_selection=gene_selection,
                         refit=refit)

    def mutate_one(self, individual, environment):
        """
        Mutates a single individual
        :param environment: the environment the individuals live in
        :param individual: the individual to mutate
        :return: None
        """
        genes_to_change = self.gene_selection(individual)

        if individual.device == 'cpu':
            individual.genes[genes_to_change] = np.random.randint(2, size=genes_to_change.size)
        if individual.device == 'gpu':
            individual.genes[genes_to_change] = cp.random.randint(2, size=genes_to_change.size)



import random

from Finch.genetics import Individual
from Finch.tools.fitness_functions import layer_fitness
from typing import Callable, Union
from Finch.tools.individualselectors import RandomSelection
from Finch.tools.geneselectors import PercentSelector, AmountSelector
from Finch.tools.rates import make_callable

cp = None
try:
    import cupy as cp
except ImportError:
    pass


class Layer(Individual):
    """
        A Layer is a building block for an Environment and modifies or sorts the individuals in some way
    """

    def __init__(self, individual_selection: Union[float, int, Callable, None] = None,
                 gene_selection: [float, int, Callable, None] = None,
                 refit=False, fitness: int = 0, device: str = 'cpu'):
        """
        :param execution_function: retrieved from a higher class
        :param gene_selection: gene selection method: float, int, Callable
        :param individual_selection: individual selection method: float, int, Callable
        :param fitness: set to 0
        :param device: 'cpu' or 'gpu'
        """
        assert device in ('cpu', 'gpu'), f"Invalid device: {self.device}. Must be either 'cpu' or 'gpu'."
        super().__init__(genes=None, fitness=fitness, device=device, gene_pool=None)  # TODO: do layers need gene pools?
        if type(individual_selection) == float:
            assert 1 >= individual_selection >= 0, f"Selection percent mut be in the range 0-1, got: {individual_selection}"
            self.individual_selection = RandomSelection(percent_to_select=individual_selection).select
        elif type(individual_selection) == int:
            self.individual_selection = RandomSelection(amount_to_select=individual_selection).select
        elif callable(individual_selection):
            self.individual_selection = individual_selection
        else:
            self.individual_selection = lambda x: x  # make it do nothing

        if type(gene_selection) == float:
            assert 1 >= gene_selection >= 0, f"Selection percent mut be in the range 0-1, got: {gene_selection}"
            self.gene_selection = PercentSelector(gene_selection).select
        elif type(gene_selection) == int:
            self.gene_selection = AmountSelector(gene_selection).select
        elif callable(gene_selection):
            self.gene_selection = gene_selection
        else:
            self.gene_selection = None

        self.execution_function = self.execute  # TODO: remove this
        self.total_fitness_contribution = 0
        self.refit = refit
        self.environment = None

    def set_environment(self, environment):
        self.environment = environment

    def run(self, individuals: list, environment):
        """
        :param individuals: The individuals in the environment
        :param environment: The environment containing the individuals
        :return: individuals
        """
        # overall fitness of the individuals before layer execution
        before = environment.get_fitness_metric()
        # run layer on selected individuals
        selected = self.individual_selection(individuals)
        self.execution_function(individuals=selected)
        # fitness on every selected individual
        if self.refit:
            for individual in selected:
                individual.fitness = environment.fitness_function(individual)
        # overall fitness after the layer executes
        after = environment.get_fitness_metric()

        # assign the layer a fitness
        self.fitness = layer_fitness(before, after)
        self.total_fitness_contribution += before - after
        return individuals

    def execute(self, individuals):
        pass


class Mutate(Layer):
    """
    Base mutation layer, not to be used directly.
    """

    def __init__(self, gene_selection, individual_selection, device='cpu', refit=True):
        """
        :param gene_selection: gene selection method
        :param individual_selection: individual selection method
        :param execution_function: execution_function (from higher layer)
        :param device: 'cpu' or 'gpu'
        """
        super().__init__(device=device, fitness=0, individual_selection=individual_selection,
                         gene_selection=gene_selection, refit=refit)

    def mutate_one(self, individual, environment):
        pass

    def execute(self, individuals):
        """
        :param individuals: individuals to execute the mutation on
        :param environment: the environment our individuals live in (:
        :return: None
        """
        for individual in individuals:
            self.mutate_one(individual, self.environment)


class Parent(Layer):
    """
    Base parenting/crossover layer, not to be used directly.
    """

    def __init__(self, families, children=1, device='cpu', refit=True, track_genealogies=False):
        """
        :param individual_selection: individual selection method
        :param execution_function:
        :param device: 'cpu' or 'gpu'
        :param track_genealogies: bool
        """
        self.families = families
        self.children = make_callable(children)
        self.track_genealogies = track_genealogies
        super().__init__(device=device, fitness=0, individual_selection=families,
                         refit=refit)

    def parent(self, parent1, parent2, environment) -> list:
        pass

    def execute(self, individuals):  # TODO: fix this all up to be good
        assert len(individuals) % 2 == 0, f"Must be an even number parents! Got: '{len(individuals)}'"
        parents = list(zip(individuals[::2], individuals[1::2]))
        for parent1, parent2 in parents:
            children = self.parent(parent1=parent1, parent2=parent2, environment=self.environment)
            if self.track_genealogies:
                for child in children:
                    child.parents.append(parent1)
                    child.parents.append(parent2)
            self.environment.individuals += children



import math

from Finch.genetics import Individual
from Finch.layers.layer import Layer
from Finch.tools.rates import make_callable
from Finch.genepools import GenePool
import numpy as np

try:
    import cupy as cp
except ImportError:
    cp = None


class Populate(Layer):
    def __init__(self, gene_pool: GenePool, population: int):
        """
        :param gene_pool: any GenePool
        :param population: amount of individuals to ensure the population has
        """
        super().__init__(individual_selection=None)
        self.gene_pool = gene_pool
        self.population = make_callable(population)

    def execute(self, individuals):
        individuals = list(individuals)  # TODO fix the need for this... Is this fixed???? I have no idea
        while len(individuals) < self.population():
            new = self.gene_pool.generate_individual()
            new.fitness = self.environment.fitness_function(new)
            individuals += [new]
        self.environment.individuals = individuals


class KillBySelection(Layer):
    def __init__(self, individual_selection):
        """
        :param individual_selection: the method by witch to select the individuals for death
        """
        super().__init__(individual_selection=individual_selection)

    def execute(self, individuals):
        # select some individuals to kill using the selection function
        # remove the selected individuals from the population
        for ind in individuals:
            ind.genes = []
            self.environment.dead_individuals.append(ind)
        individuals = [ind for ind in self.environment.individuals if ind not in individuals]
        self.environment.individuals = individuals


class DuplicateSelection(Layer):
    def __init__(self, individual_selection):
        """
        :param individual_selection:
        """
        super().__init__(individual_selection=individual_selection)

    def execute(self, individuals):
        # select some individuals to duplicate using the selection function

        duplicates = [d.copy() for d in individuals]  # make them different (:
        individuals += duplicates


class RemoveDuplicatesFromTop(Layer):
    def __init__(self, top_n: int):
        super().__init__(individual_selection=None)
        self.top_n = make_callable(top_n)

    def execute(self, individuals):
        # remove any duplicates from the top n individuals
        unique_individuals = individuals[:self.top_n()]
        if self.environment.device == 'cpu':
            npcp = np
        else:
            npcp = cp

        for i in range(self.top_n, len(individuals)):
            if not any(npcp.array_equal(ind.genes, individuals[i].genes) for ind in unique_individuals):
                unique_individuals.append(individuals[i])
        return unique_individuals


class SortByFitness(Layer):
    def __init__(self):
        super().__init__(individual_selection=None)

    def execute(self, individuals):
        sorted_individuals = sorted(individuals, key=lambda x: -x.fitness)
        self.environment.individuals = list(sorted_individuals)


class CapPopulation(Layer):
    def __init__(self, max_population: int):
        super().__init__(individual_selection=None)
        self.max_population = make_callable(max_population)

    def execute(self, individuals):
        self.environment.individuals = individuals[
                                       0:self.max_population()]  # kills only the worst ones assuming they are sorted

        for i in individuals[self.max_population():-1]:
            i.genes = []
            self.environment.dead_individuals.append(i)


class BatchFitness(Layer):
    """
    Layer applying a batch fitness function to selected individuals.

    Parameters:
    - batch_fitness_function: Callable, batch fitness function to be applied.

    Attributes:
    - batch_fitness_function: Callable, the batch fitness function.

    Methods:
    - execute(individuals: List[Individual]) -> None:
        Apply batch fitness function to selected individuals.
    """

    def __init__(self, batch_fitness_function):
        super().__init__()
        self.batch_fitness_function = batch_fitness_function

    def execute(self, individuals: list) -> None:
        """
        Apply batch fitness function to selected individuals.

        Parameters:
        - individuals: List[Individual], individuals to process.

        Returns:
        - None
        """
        selected = [individual for individual in individuals if individual.check_fitness]
        self.batch_fitness_function(selected)
        for i in selected:
            i.check_fitness = False


class Function(Layer):
    def __init__(self, function, individual_selection=None):
        super().__init__(individual_selection=individual_selection)
        self.function = function

    def execute(self, individuals):
        return self.function(individuals)


class Controller(Layer):
    def __init__(self, layer: Layer, execute_every=1, repeat=1, delay=0, stop_at=math.inf):
        super().__init__(individual_selection=None)
        self.layer = layer
        self.every = make_callable(execute_every)
        self.delay = make_callable(delay)
        self.end = stop_at
        self.repeat = make_callable(repeat)
        self.n = 0

    def execute(self, individuals):
        self.n += 1
        if self.delay() <= self.n <= self.end:
            if self.every() % self.n == 0:
                for i in range(self.repeat() + 1):
                    individuals = self.layer.execute(individuals)
        return individuals


class RemoveAllButBest(Layer):
    def __init__(self):
        super().__init__(individual_selection=None)

    def execute(self, individuals):
        for individual in individuals[1:-1]:
            individual.genes = [] # saves on memory, but retains everything else important about individuals
            self.environment.dead_individuals.append(individual)
        self.environment.individuals = [individuals[0]]


#
# class FreezeRandom(Layer):
#     def __init__(self, amount_genes: int, selection_function: callable = randomSelect):
#         super().__init__(individual_selection=individual_selection)
#         self.amount_genes = make_callable(amount_genes)
#
#     def execute(self, individuals):
#         selected_individuals = self.selection_function.select(individuals)
#         if environment.device == 'cpu':
#             npcp = np
#         else:
#             npcp = cp
#         for individual in selected_individuals:
#             random_indices = npcp.random.choice(individual.genes.size, self.amount_genes(), replace=False)
#             individual.freeze(random_indices)
#         return individuals





"""
Something here
"""
import numpy as np

cp = None
try:
    import cupy as cp
except ImportError:
    pass


class Individual:
    """
    Represents an individual in a population.

    This class can be used to model various types of individuals, such as images, books, or vectors.

    Attributes:
        genes (Any): The genetic information or characteristics of the individual.
        fitness (float): The default fitness value for this individual.
        parents (list): The parents of the individual
        children (list): Any children the parents have had.
    Example:
        >>> individual = Individual(genes=[0, 1, 1, 0], fitness=0.75)
    """

    def __init__(self, gene_pool, genes: np.array, fitness: float = 0, device="cpu"):
        self.genes = genes
        self.fitness = fitness
        self.device = device
        self.gene_pool = gene_pool
        self.check_fitness = False

        self.children = []
        self.parents = []
        self.age = 0

    def copy(self):
        """
        :return: A new copy of the individual
        """
        if self.device == "gpu":
            return Individual(genes=cp.copy(self.genes), fitness=self.fitness, device=self.device,
                              gene_pool=self.gene_pool)
        else:
            return Individual(genes=np.copy(self.genes), fitness=self.fitness, device=self.device,
                              gene_pool=self.gene_pool)

    def to_cpu(self):
        """
        self.genes will move to cpu and use numpy
        :return: self
        """
        if self.device == "gpu":
            self.genes = cp.asnumpy(self.genes)
            self.device = "cpu"
        return self

    def to_gpu(self):
        """
        self.genes will move to gpu and use cupy
        :return: self
        """
        if self.device == "cpu":
            self.genes = cp.asarray(self.genes)
            self.device = "gpu"
        return self



import random
from Finch.genetics import Individual
from Finch.tools.rates import make_callable


class Select:
    """
    Base class for selection strategies.

    Parameters:
    - percent_to_select: A callable returning the percentage of individuals to select.
    - amount_to_select: A callable returning the number of individuals to select.

    Usage:
    ```
    selector = Select(percent_to_select=lambda: 0.2)
    selected_individuals = selector.select(individuals)
    ```
    """

    def __init__(self, percent_to_select=None, amount_to_select=None):
        if percent_to_select is not None and amount_to_select is not None:
            raise ValueError("Only one of percent_to_select or amount_to_select can be given")

        self.percent_to_select = make_callable(percent_to_select)
        self.amount_to_select = make_callable(amount_to_select)

    def select(self, individuals: list[Individual]):
        """
        Abstract method for selecting individuals.

        Parameters:
        - individuals: List of individuals to select from.

        Returns:
        - list[Individual]: Selected individuals.
        """
        pass


class TournamentSelection(Select):
    """
    Tournament selection strategy.

    Parameters:
    - percent_to_select: A callable returning the percentage of individuals to select.
    - amount_to_select: A callable returning the number of individuals to select.

    Usage:
    ```
    selector = TournamentSelection(percent_to_select=lambda: 0.2)
    selected_individuals = selector.select(individuals)
    ```
    """

    def __init__(self, percent_to_select=None, amount_to_select=None):
        super().__init__(percent_to_select, amount_to_select)

    def select(self, individuals):
        """
        Select individuals using tournament selection.

        Parameters:
        - individuals: List of individuals to select from.

        Returns:
        - list[Individual]: Selected individuals.
        """
        selected_individuals = []

        if self.percent_to_select is not None:
            amount = int(self.percent_to_select() * len(individuals))
        else:
            amount = self.amount_to_select()

        for _ in range(amount):
            tournament_size = len(individuals)
            tournament_individuals = random.sample(individuals, k=tournament_size)
            winner = max(tournament_individuals, key=lambda individual: individual.fitness)
            selected_individuals.append(winner)

        return selected_individuals


class RandomSelection(Select):
    """
    Random selection strategy.

    Parameters:
    - percent_to_select: A callable returning the percentage of individuals to select.
    - amount_to_select: A callable returning the number of individuals to select.

    Usage:
    ```
    selector = RandomSelection(percent_to_select=lambda: 0.2)
    selected_individuals = selector.select(individuals)
    ```
    """

    def __init__(self, percent_to_select=None, amount_to_select=None):
        super().__init__(percent_to_select, amount_to_select)

    def select(self, individuals):
        """
        Select individuals randomly.

        Parameters:
        - individuals: List of individuals to select from.

        Returns:
        - list[Individual]: Selected individuals.
        """
        if self.percent_to_select is not None:
            amount = int(self.percent_to_select() * len(individuals))
        else:
            amount = self.amount_to_select()
        selected_individuals = random.choices(individuals, k=amount)
        return selected_individuals


class RankBasedSelection(Select):
    """
    Rank-based selection strategy.

    Parameters:
    - factor: Selection pressure factor.
    - percent_to_select: A callable returning the percentage of individuals to select.
    - amount_to_select: A callable returning the number of individuals to select.

    Usage:
    ```
    selector = RankBasedSelection(factor=1.5, percent_to_select=lambda: 0.2)
    selected_individuals = selector.select(individuals)
    ```
    """

    def __init__(self, factor, percent_to_select=None, amount_to_select=None):
        super().__init__(percent_to_select, amount_to_select)
        self.factor = factor

    def select(self, individuals):
        """
        Select individuals using rank-based selection.

        Parameters:
        - individuals: List of individuals to select from.

        Returns:
        - list[Individual]: Selected individuals.
        """
        population_size = len(individuals)
        ranks = list(range(1, population_size + 1))

        selection_probs = [pow(2.71828, -self.factor * rank / population_size) for rank in ranks]

        sum_probs = sum(selection_probs)
        selection_probs = [prob / sum_probs for prob in selection_probs]

        if self.percent_to_select is not None:
            amount = int(self.percent_to_select() * len(individuals))
        else:
            amount = self.amount_to_select()

        selected_indices = random.choices(range(population_size), weights=selection_probs, k=amount)
        selected_individuals = [individuals[i] for i in selected_indices]

        return selected_individuals



import numpy as np
from Finch.genetics import Individual
from Finch.layers import layer
import random

cp = None
try:
    import cupy as cp
except ImportError:
    pass


def crossover(parent1: Individual, parent2: Individual) -> list:
    """
    Perform single-point crossover between two parents.

    Args:
        parent1 (Individual): The first parent.
        parent2 (Individual): The second parent.

    Returns:
        list of Individual: Two offspring resulting from the crossover.
    """
    average_fitness = (parent1.fitness + parent2.fitness) / 2
    point = random.randint(1, parent1.genes.size - 1)
    array_manager = cp if parent1.device == 'gpu' else np

    offspring1 = Individual(genes=array_manager.append(parent1.genes[:point], parent2.genes[point:]),
                            fitness=average_fitness,
                            device=parent1.device, gene_pool=parent1.gene_pool)

    offspring2 = Individual(genes=array_manager.append(parent2.genes[:point], parent1.genes[point:]),
                            fitness=average_fitness,
                            device=parent1.device, gene_pool=parent1.gene_pool)

    return [offspring1, offspring2]


def uniform_crossover(parent1: Individual, parent2: Individual, probability=0.5) -> list:
    """
    Perform uniform crossover between two parents.

    Args:
        parent1 (Individual): The first parent.
        parent2 (Individual): The second parent.
        probability (float): The probability of selecting a gene from the first parent.

    Returns:
        list of Individual: Two offspring resulting from the crossover.
    """
    average_fitness = (parent1.fitness + parent2.fitness) / 2
    array_manager = cp if parent1.device == 'gpu' else np

    mask = array_manager.random.uniform(0, 1, size=parent1.genes.size) < probability

    genes1 = array_manager.where(mask, parent1.genes, parent2.genes)
    genes2 = array_manager.where(mask, parent2.genes, parent1.genes)

    offspring1 = Individual(genes=genes1,
                            fitness=average_fitness,
                            device=parent1.device, gene_pool=parent1.gene_pool)

    offspring2 = Individual(genes=genes2,
                            fitness=average_fitness,
                            device=parent1.device, gene_pool=parent1.gene_pool)

    return [offspring1, offspring2]


def n_point_crossover(parent1: Individual, parent2: Individual, n=1) -> list:
    """
    Perform n-point crossover between two parents.

    Args:
        parent1 (Individual): The first parent.
        parent2 (Individual): The second parent.
        n (int): Number of crossover points.

    Returns:
        list of Individual: Offspring resulting from the crossover.
    """
    average_fitness = (parent1.fitness + parent2.fitness) / 2
    gene_length = parent1.genes.size
    array_manager = cp if parent1.device == 'gpu' else np

    points = sorted(random.sample(range(1, gene_length), n))
    points.append(gene_length)

    mask = array_manager.zeros(gene_length, dtype=bool)
    current_parent = 1  # Start with parent 1

    for point in points:
        mask[point:] = ~mask[point:]
        current_parent = 3 - current_parent  # Switch parent

    genes1 = array_manager.where(mask, parent1.genes, parent2.genes)

    offspring = Individual(genes=genes1,
                           fitness=average_fitness,
                           device=parent1.device, gene_pool=parent1.gene_pool)

    return [offspring]


class ParentSimple(layer.Parent):
    def __init__(self, families, children=2, refit=True):
        """
        Initialize a simple parent with crossover.

        Args:
            families: Selection method for individuals.
            children (int): Number of children to generate.
            refit (bool): If true will retest fitness in new children
        """
        if type(children) == int:
            assert children > 1, f"This type of crossover must return more than 1 child, you specified '{children}'."
        super().__init__(families=families,  children=children, refit=refit)

    def parent(self, parent1: Individual, parent2: Individual, environment) -> list:
        """
        Generate children through crossover.

        Args:
            parent1 (Individual): The first parent.
            parent2 (Individual): The second parent.
            environment: Additional environment information.

        Returns:
            list of Individual: Offspring resulting from the crossover.
        """
        return crossover(parent1=parent1, parent2=parent2)


class ParentNPoint(layer.Parent):
    def __init__(self, families, points=1, children=2, refit=True, track_genealogies = False):
        """
        Initialize an n-point parent with crossover.

        Args:
            families: Selection method for individuals.
            points (int): Number of crossover points.
            children (int): Number of children to generate.
            refit (bool): If true will retest fitness in new children
            track_genealogies (bool): If true will have individuals remember their parents
        """
        self.points = points
        super().__init__(families=families,  children=children, refit=refit, track_genealogies=track_genealogies)

    def parent(self, parent1: Individual, parent2: Individual, environment) -> list:
        """
        Generate children through n-point crossover.

        Args:
            parent1 (Individual): The first parent.
            parent2 (Individual): The second parent.
            environment: Additional environment information.

        Returns:
            list of Individual: Offspring resulting from the crossover.
        """

        children =  n_point_crossover(parent1=parent1, parent2=parent2, n=self.points)
        return children



import random
import matplotlib.pyplot as plt
from typing import Union


class Rate:
    def __init__(self, start, end, epochs, return_int=False):
        """
        :param start: The starting value (int or float)
        :param end: The target value after epochs
        :param epochs: The number of epochs to reach the target value
        :param return_int: Whether to return only integers
        """
        self.start = start
        self.value = start  # use a more descriptive name than start
        self.end = end
        self.epochs = epochs
        self.change_rate = (end - start) / epochs  # use underscores for variable names
        self.return_int = return_int  # use the same name as the parameter

    def next(self):
        # return the current value and update it by the change rate
        if self.value < self.end < self.start:
            return self.end
        if self.value > self.end > self.start:
            return self.end
        result = self.value
        self.value += self.change_rate

        if self.return_int:
            return int(result)
        return result

    def get(self):
        # return the current value without updating it
        if self.return_int:
            return int(self.value)
        else:
            return self.value

    def graph(self):
        # plot the history of the value over epochs
        history = []
        temp = self.value  # store the current value temporarily
        for i in range(self.epochs):
            history.append(temp)
            temp += self.change_rate
        print("min", min(history))
        print("max", max(history))
        plt.plot(history)
        plt.show()
        self.value = temp  # restore the current value


def make_switcher(x) -> callable(any):
    x = make_callable(x)
    n = x()

    def r():
        # return a random choice of -n or n
        return random.choice([-n, n])

    return r


def make_callable(x: Union[callable, int, float]) -> callable:
    if x is None:
        return None
    if not callable(x):
        # return a function that always returns x
        def constant() -> int | float:
            return x

        return constant
    else:
        return x



from setuptools import setup, find_packages

with open('Finch/README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()

setup(
    name='finch-genetics',
    version='3.0.5',
    packages=find_packages(),
    install_requires=[
        "torch",
        "transformers",
        "diffusers",
        "numpy",
        "matplotlib",
        "typing_extensions"
    ],
    author='Daniel Losey',
    author_email='danieljlosey@gmail.com',
    description='GPU and CPU modular genetic algorithm framework.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/dadukhankevin/Finch',
    classifiers=[
        'Programming Language :: Python :: 3',
        # Add other classifiers as needed
    ],
)



import numpy as np
import torch
from transformers import pipeline
from PIL import Image
from Finch.genetics import Individual
from diffusers import AutoPipelineForText2Image

cp = None
try:
    import cupy as cp
except ImportError:
    pass

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


class ZeroShotImage:
    def __init__(self, target_labels, other_labels, shape, model="openai/clip-vit-large-patch14", criteria=sum,
                 base_image=None, denormalize=False, batch_size=1):
        self.model = model
        self.target_labels = target_labels
        self.other_labels = other_labels
        self.pipe = pipeline("zero-shot-image-classification", model=model, device=device)
        self.criteria = criteria
        self.base_image = base_image
        self.shape = shape
        self.size = (shape[0], shape[1])
        self.denormalize = denormalize
        self.batch_size = batch_size

    def run(self, image, candidate_labels):
        return self.pipe(image, candidate_labels=candidate_labels, batch_size=self.batch_size)

    def search(self, terms, scores):
        return self.criteria([term['score'] for term in scores if term['label'] in terms])

    def suppress_fit(self, individual: Individual):
        if individual.device == 'gpu':
            image_array = cp.asnumpy(individual.genes)
        else:
            image_array = individual.genes
        if self.denormalize:
            image_array *= 255
        if self.base_image:
            image_array += self.base_image

        image_array = image_array.reshape(self.shape).astype(np.uint8)

        image = Image.fromarray(image_array)

        return self.supress_pil_image(image)

    def supress_pil_image(self, image: Image):
        labels = self.run(image, self.target_labels + self.other_labels)
        return -self.search(self.target_labels, labels)

    def enhance_pil_image(self, image: Image):
        return -self.supress_pil_image(image)

    def enhance_fit(self, individual: Individual):
        return -self.suppress_fit(individual)

    def get_image(self, individual):
        if individual.device == 'gpu':
            image_array = cp.asnumpy(individual.genes)
        else:
            image_array = individual.genes
        if self.denormalize:
            image_array *= 255
        if self.base_image:
            image_array += self.base_image

        image_array = image_array.reshape(self.shape).astype(np.uint8)
        image = Image.fromarray(image_array)
        return image

    def batch_enhance(self, individuals, pil_images=False):
        if not pil_images:
            raw_images = [self.get_image(im) for im in individuals]
        else:
            raw_images = pil_images

        scores = self.run(raw_images, candidate_labels=self.target_labels + self.other_labels)
        ind = 0
        for score in scores:
            individuals[ind].fitness = self.search(self.target_labels, score)
            ind += 1

    def batch_supress(self, individuals, pil_images):
        self.batch_enhance(individuals, pil_images)
        for individual in individuals:
            individual.fitness *= -1

    def show(self, individual):
        image = self.get_image(individual)
        image.show()
        return image


class ObjectDetection:
    def __init__(self, target_labels, other_labels, shape, model="hustvl/yolos-tiny", criteria=sum,
                 base_image=None, denormalize=False, batch_size=1):
        # Use meaningful names for the parameters
        self.model_name = model
        self.target_labels = target_labels
        self.other_labels = other_labels
        self.shape = shape
        self.criteria = criteria
        self.base_image = base_image
        self.denormalize = denormalize
        self.batch_size = batch_size

        # Use the built-in pipeline function to create the object detector
        self.pipe = pipeline("object-detection", model=self.model_name, device=device)

    def run(self, image):
        # Use the pipeline to get the labels and scores for the image
        return self.pipe(image)

    def search(self, terms, scores):
        # Use a list comprehension to get the scores for the terms
        return self.criteria([term['score'] for term in scores if term['label'] in terms])

    def suppress_fit(self, individual: Individual):
        # Use numpy to convert the genes to a numpy array
        image_array = np.array(individual.genes)

        # Use the denormalize and base_image attributes to modify the image array
        if self.denormalize:
            image_array *= 255
        if self.base_image:
            image_array += self.base_image

        # Reshape and cast the image array to uint8
        image_array = image_array.reshape(self.shape).astype(np.uint8)

        # Create an image from the array
        image = Image.fromarray(image_array)

        # Get the labels and scores for the image
        labels = self.run(image)

        # Return the negative score for the target labels
        return -self.search(self.target_labels, labels)

    def supress_pil_image(self, image: Image):
        # Return the negative score for the target labels
        return self.suppress_fit(image)

    def enhance_pil_image(self, image: Image):
        # Return the positive score for the target labels
        return -self.supress_pil_image(image)

    def enhance_fit(self, individual: Individual):
        # Return the positive score for the target labels
        return -self.suppress_fit(individual)

    def get_image(self, individual):
        # Use numpy to convert the genes to a numpy array
        image_array = np.array(individual.genes)

        # Use the denormalize and base_image attributes to modify the image array
        if self.denormalize:
            image_array *= 255
        if self.base_image:
            image_array += self.base_image

        # Reshape and cast the image array to uint8
        image_array = image_array.reshape(self.shape).astype(np.uint8)

        # Create and return an image from the array
        return Image.fromarray(image_array)

    def batch_enhance(self, individuals, pil_images=False):
        # Use a list comprehension to get the images from the individuals
        if not pil_images:
            raw_images = [self.get_image(im) for im in individuals]
        else:
            raw_images = pil_images

        # Get the labels and scores for the images
        scores = self.run(raw_images)

        # Use a for loop with enumerate to update the fitness of the individuals
        for ind, score in enumerate(scores):
            individuals[ind].fitness = self.search(self.target_labels, score)

    def batch_supress(self, individuals, pil_images):
        # Use the batch_enhance method to get the fitness of the individuals
        self.batch_enhance(individuals, pil_images)

        # Use a list comprehension to negate the fitness of the individuals
        for individual in individuals:
            individual.fitness *= -1

    def show(self, individual):
        # Use the get_image method to get the image from the individual
        image = self.get_image(individual)

        # Show the image
        image.show()

        # Return the image
        return image


class ImageGenerator:
    def __init__(self, recognizer: ZeroShotImage, model='stabilityai/sd-turbo', variant='fp16',
                 batch_size=1, guidance_scale=0.0, num_inference_steps=1, repeat_fitness=1, criteria=sum, seed=-1):
        self.pipe = AutoPipelineForText2Image.from_pretrained(model, torch_dtype=torch.float16,
                                                              variant=variant)
        self.pipe.to("cuda")
        self.batch_size = batch_size
        self.recognizer = recognizer
        self.guidance_scale = guidance_scale
        self.num_inference_steps = num_inference_steps
        self.repeat_fitness = repeat_fitness
        self.criteria = criteria
        self.seed = seed

    def generate(self, prompt):
        image = self.pipe(prompt=prompt, num_inference_steps=self.num_inference_steps,
                          guidance_scale=self.guidance_scale,
                          batch_size=self.batch_size, seed=self.seed, size=self.recognizer.size)[0]
        return image

    def fit(self, individual: Individual):
        prompt = "".join(str(i) for i in individual.genes)
        image = self.generate(prompt)[0]
        return self.recognizer.enhance_pil_image(image)

    def repeated_fit(self, individual: Individual):
        assert self.seed == -1, "Repeated_fit should only be used when seed is randomized, please set it to -1"
        return self.criteria([self.fit(individual) for _ in range(self.repeat_fitness)])

    def batch_fit(self, individuals: list[Individual]):
        prompts = ["".join(str(i) for i in individual.genes) for individual in individuals]
        images = self.generate(prompts)
        self.recognizer.batch_enhance(individuals, pil_images=images)

    def batch_repeated_fit(self, individuals):
        fitness_values_list = []

        for _ in range(self.repeat_fitness):
            self.batch_fit(individuals)
            fitness_values = [individual.fitness for individual in individuals]
            fitness_values_list.append(fitness_values)

        aggregated_fitness_values = [self.criteria(values) for values in zip(*fitness_values_list)]

        for i, individual in enumerate(individuals):
            individual.fitness = aggregated_fitness_values[i]


from typing import Union, Callable
import numpy as np
from Finch.genetics import Individual

cp = None
try:
    import cupy as cp
except ImportError:
    pass

class GeneSelector:
    def __init__(self):
        """
        Base class for gene selectors.
        """
        pass

    def select(self, individual: Individual) -> np.ndarray:
        """
        Abstract method to be implemented by subclasses.

        Parameters:
        - individual: An instance of Individual from which genes will be selected.

        Returns:
        - np.ndarray: An array containing the selected genes.
        """
        pass


class PercentSelector(GeneSelector):
    def __init__(self, percent: Union[float, int, Callable]):
        """
        Initialize PercentSelector with the specified percentage.

        Parameters:
        - percent: The percentage of genes to be selected. It can be a float, int, or a callable (function) that takes an individual as an argument.
        """
        super().__init__()
        self.percent = percent

    def select(self, individual: Individual) -> np.ndarray:
        """
        Select a certain percentage of genes from an individual.

        Parameters:
        - individual: An instance of Individual from which genes will be selected.

        Returns:
        - np.ndarray: An array containing the indices of the selected genes.
        """
        device = individual.device
        genes = individual.genes

        if callable(self.percent):
            percent = self.percent(individual)
        else:
            percent = float(self.percent)

        amount = int(len(genes) * percent)
        indices = np.arange(len(genes))

        if device == 'gpu':
            selected_indices = cp.random.choice(indices, size=amount, replace=True)
        elif device == 'cpu':
            selected_indices = np.random.choice(indices, size=amount, replace=True)

        return selected_indices


class AmountSelector(GeneSelector):
    def __init__(self, amount: Union[int, Callable]):
        """
        Initialize AmountSelector with the specified number of genes to be selected.

        Parameters:
        - amount: The number of genes to be selected. It can be an int or a callable (function) that takes an individual as an argument.
        """
        super().__init__()
        self.amount = amount

    def select(self, individual: Individual) -> np.ndarray:
        """
        Select a specified number of genes from an individual.

        Parameters:
        - individual: An instance of Individual from which genes will be selected.

        Returns:
        - np.ndarray: An array containing the indices of the selected genes.
        """
        if callable(self.amount):
            amount = self.amount()
        else:
            amount = self.amount
        device = individual.device
        genes = individual.genes
        indices = np.arange(len(genes))

        if device == 'gpu':
            selected_indices = cp.random.choice(indices, size=amount, replace=True)
        elif device == 'cpu':
            selected_indices = np.random.choice(indices, size=amount, replace=True)

        return selected_indices

# Example usage:
# individual = Individual(genes=np.arange(10), device='cpu')
# selector = PercentSelector(percent=0.3)
# selected_indices = selector.select(individual)
# print(selected_indices)



"""
GenePools

This script defines various gene pool classes for the creation of individuals with different types of genes.
Each gene pool class inherits from the base class GenePool and implements methods for generating individuals
with specific gene types.

Classes:
- GenePool: Base class for gene pools. (not usable)
- FloatPool: Gene pool for individuals containing floating-point genes.
- IntPool: Gene pool for individuals containing integer genes.
- BinaryPool: Gene pool for individuals containing binary genes (0 or 1).
- ArrayPool: Gene pool for individuals by picking items from an array.

Author: Daniel Losey
Year: 2023
"""
from Finch import genetics
import numpy as np

cp = None
try:
    import cupy as cp
except ImportError:
    pass


class GenePool:
    def __init__(self, length: int, device='cpu'):
        """
        Base GenePool class
        :param length: Amount of genes to place in an individual
        :param device: 'cpu' or 'gpu'
        """
        self.device = device
        self.length = length
        assert self.device in ('cpu', 'gpu'), f"Invalid device: {self.device}. Must be either 'cpu' or 'gpu'."

    def generate_individual(self):
        """
        Generates a new individual.
        :return: genetics.Individual with genes of type int
        """
        genes = self.generate_genes(self.length)
        return genetics.Individual(genes=genes, fitness=0, device=self.device, gene_pool=self)

    def generate_genes(self, amount: int):
        pass


class FloatPool(GenePool):
    def __init__(self, length: int, minimum: float = 0, maximum: float = 1, default: float = None, device="cpu"):
        """
        A GenePool meant specifically for the creation of individuals containing floats.
        :param length: Amount of genes in each individual
        :param minimum: Minimum value for a gene
        :param maximum: Maximum value for a gene
        :param default: If set will initialize all genes to the same value
        :param device: Where genes in Individuals should be kept 'gpu' or 'cpu'
        """
        super().__init__(length=length, device=device)
        self.minimum = minimum
        self.maximum = maximum
        self.default = default

    def generate_genes(self, amount: int):
        """
        :param amount: Amount of genes to generate
        :return: numpy or cupy array containing genes
        """
        if self.device == "cpu":  # generate genes using numpy
            if self.default is not None:
                genes = np.full(amount, self.default)
            else:
                genes = np.random.uniform(self.minimum, self.maximum, size=amount)

        elif self.device == "gpu":  # generate genes using cupy
            if self.default is not None:
                genes = cp.full(amount, self.default)
            else:
                genes = cp.random.uniform(self.minimum, self.maximum, size=amount)

        return genes


class IntPool(GenePool):
    def __init__(self, length: int, minimum: int = 0, maximum: int = 1, default: int = None, device="cpu"):
        """
        A GenePool meant specifically for the creation of individuals containing floats.
        :param length: Amount of genes in each individual
        :param minimum: Minimum value for a gene
        :param maximum: Maximum value for a gene
        :param default: If set will initialize all genes to the same value
        :param device: Where genes in Individuals should be kept 'gpu' or 'cpu'
        """
        super().__init__(length=length, device=device)
        self.minimum = minimum
        self.maximum = maximum
        self.default = default

    def generate_genes(self, amount: int):
        """
        :param amount: Amount of genes to generate
        :return: numpy or cupy array containing genes
        """
        if self.device == "cpu":  # generate genes using numpy
            if self.default is not None:
                genes = np.full(amount, self.default)
            else:
                genes = np.random.randint(self.minimum, self.maximum + 1, size=amount)

        elif self.device == "gpu":  # generate genes using cupy
            if self.default is not None:
                genes = cp.full(amount, self.default)
            else:
                genes = cp.random.randint(self.minimum, self.maximum + 1, size=amount)
        return genes


class BinaryPool(GenePool):
    def __init__(self, length: int, default: int = None, device="cpu"):
        """
        A GenePool meant specifically for the creation of individuals containing binary genes (0 or 1).
        :param length: Amount of genes in each individual
        :param default: If set will initialize all genes to the same value
        :param device: Where genes in Individuals should be kept 'gpu' or 'cpu'
        """
        super().__init__(length=length, device=device)
        self.default = default

    def generate_genes(self, amount: int):
        """
        :param amount: Amount of genes to generate
        :return: numpy or cupy array containing binary genes (0 or 1)
        """
        if self.device == "cpu":  # generate genes using numpy
            if self.default is not None:
                genes = np.full(amount, self.default, dtype=int)
            else:
                genes = np.random.randint(2, size=amount, dtype=int)

        elif self.device == "gpu":  # generate genes using cupy
            if self.default is not None:
                genes = cp.full(amount, self.default, dtype=int)
            else:
                genes = cp.random.randint(2, size=amount, dtype=int)

        return genes


class ArrayPool(GenePool):
    def __init__(self, gene_array: np.ndarray, length: int, device="cpu"):
        """
        A GenePool meant for the creation of individuals by picking items from an array.
        :param gene_array: The array from which genes will be picked.
        :param length: Amount of genes in each individual
        :param device: Where genes in Individuals should be kept 'gpu' or 'cpu'
        """
        super().__init__(length=length, device=device)
        self.gene_array = gene_array

    def generate_genes(self, amount: int):
        """
        :param amount: Amount of genes to generate
        :return: numpy or cupy array containing genes picked from the array
        """
        if self.device == "cpu":
            genes = np.random.choice(self.gene_array, size=amount)

        elif self.device == "gpu":
            genes = cp.random.choice(self.gene_array, size=amount)
        return genes



from typing import Union
from Finch.tools.individualselectors import RankBasedSelection


class Fitness:
    def __init__(self):
        """Initialize a generic Fitness object."""
        pass

    def fit(self, individual):
        """Placeholder method for applying fitness to an individual."""
        pass


class FitnessChain(Fitness):
    def __init__(self, functions: list[tuple[callable, Union[float, int]]]):
        """
        Initialize a FitnessChain object.

        Parameters:
        - functions: List of tuples with a callable fitness function and its associated threshold.
        """
        super().__init__()
        self.functions = functions
        self.index = 0

    def fit(self, individual):
        """Apply the current fitness function to an individual."""
        self.functions[self.index][0](individual)

    def callback(self, environment):
        """Update the index if the current individual's fitness exceeds the threshold."""
        if environment.individuals[-1].fitness > self.functions[self.index][1]:
            self.index += 1


class MixFitnessByFactor(Fitness):
    """
    Combine fitness functions with weights (summing up to one) to calculate individual fitness.
    """

    def __init__(self, functions, weights, track_fitness=False):
        """
        Initialize a MixFitness object.

        Parameters:
        - functions: List of fitness functions to be combined.
        - weights: List of weights corresponding to each fitness function (must add up to one).
        - track_fitness: Bool, if true will track each functions fitness histories
        """
        super().__init__()
        self.functions = functions
        self.weights = weights
        self.track_fitness = track_fitness
        self.history = [[0] * len(functions)]

    def fit(self, individual):
        """
        Calculate the combined fitness of an individual using provided functions and weights.

        Parameters:
        - individual: The individual for which combined fitness is calculated.

        Returns:
        - float: The combined fitness value.
        """
        if len(self.functions) != len(self.weights):
            raise ValueError("Number of functions must be equal to the number of weights.")

        total_fitness = 0.0
        ind = 0
        for func, weight in zip(self.functions, self.weights):
            new_fitness = weight * func(individual)
            if self.track_fitness:
                self.history[ind].append(new_fitness)
            total_fitness += new_fitness
            ind += 1

        return total_fitness


class MixFitness(Fitness):
    def __init__(self, functions, criteria=min, track_fitness=False):
        """
        Initialize a MixFitness object.

        Parameters:
        - functions: List of fitness functions to be combined.
        - criteria: The aggregation criteria to combine fitness values, either `min` or `max`.
                    Defaults to `min`.
        - track_fitness: Bool, if true will track each function's fitness histories.
        """
        super().__init__()
        self.functions = functions
        self.criteria = criteria
        self.track_fitness = track_fitness
        self.history = [[] for _ in range(len(functions))] if track_fitness else None

    def fit(self, individual):
        """
        Calculate the combined fitness of an individual using provided functions and aggregation criteria.

        Parameters:
        - individual: The individual for which combined fitness is calculated.

        Returns:
        - float: The combined fitness value.
        """
        scores = [function(individual) for function in self.functions]

        if self.track_fitness:
            for i, score in enumerate(scores):
                self.history[i].append(score)

        return self.criteria(scores)


